﻿// Example of different ways to do a parallel calculation in C# and .NET.
//
//Test results on a 6-core desktop:
//% ParallelCompare.exe 10000000 2000
//Serial, ms = 95087
//
//Classic threaded, 2 threads, ms = 49770
//Classic threaded, 3 threads, ms = 35578
//Classic threaded, 4 threads, ms = 27966
//Classic threaded, 5 threads, ms = 20876
//Classic threaded, 6 threads, ms = 18633
//Classic threaded, 7 threads, ms = 22096
//Classic threaded, 8 threads, ms = 20463
//
//Thread pool, 2 threads, ms = 48565
//Thread pool, 3 threads, ms = 32897
//Thread pool, 4 threads, ms = 25827
//Thread pool, 5 threads, ms = 20694
//Thread pool, 6 threads, ms = 17457
//Thread pool, 7 threads, ms = 21873
//Thread pool, 8 threads, ms = 21557
//
//Thread tasks, 2 threads, ms = 52290
//Thread tasks, 3 threads, ms = 35530
//Thread tasks, 4 threads, ms = 27627
//Thread tasks, 5 threads, ms = 22222
//Thread tasks, 6 threads, ms = 18720
//Thread tasks, 7 threads, ms = 22935
//Thread tasks, 8 threads, ms = 20464
//
//Parallel for, 2 threads, ms = 52379
//Parallel for, 3 threads, ms = 35403
//Parallel for, 4 threads, ms = 27597
//Parallel for, 5 threads, ms = 22171
//Parallel for, 6 threads, ms = 18534
//Parallel for, 7 threads, ms = 22842
//Parallel for, 8 threads, ms = 20521

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;
using System.Threading;

namespace ParallelCompare
{
    class ParallelCompare
    {
        private static int countFinished = 0;

        // Calculate how often a set of target values occur in a big array of data.
        // Puts answer into slots in result array, which caller must have allocated.
        // Starts with target value start and advanced by stride each iteration.
        public static void count(int[] data, int[] targets, int [] result, int start, int stride)
        {
            //System.Console.WriteLine("count, start = {0}, stride = {1}", start, stride);
            for (int i = 0; i < data.Length; i++)
            {
                for (int j = start; j < targets.Length; j += stride)
                {
                    if (data[i] == targets[j])
                        result[j]++;
                }
            }
            Interlocked.Increment(ref countFinished);
        }

        // Launch a bunch of threads to do the comparison in parallel.
        // A classic approach to parallelism, launch threads and wait for them to do their work.
        // Makes use of an anonymous method via a lambda expression (.NET v3.5).
        public static void countThreaded(int[] data, int[] targets, int[] result, int numThreads)
        {
            List<Thread> threads = new List<Thread>();
            for (int i = 0; i < numThreads; i++)
            {
                int start = i;  // note this is required, using i in the lambda expresssion fails (?!?)
                Thread t = new Thread(() => count(data, targets, result, start, numThreads));
                // t.Priority = ThreadPriority.BelowNormal;  // helps keep Pandora from skipping when >= 4 threads on my Macbook
                threads.Add(t);
                t.Start();
            }
            
            // Wait for all the threads to complete, doesn't matter the order since we need them all to finish
            foreach (Thread t in threads)
                t.Join();
        }

        // Launch a bunch of threads to do the comparison in parallel.
        // This version uses the ThreadPool to handle the jobs.
        // ThreadPool has been around since .NET v1.0.
        public static void countPool(int[] data, int[] targets, int[] result, int numThreads)
        {
            // We need a way to know when all the threads are done
            countFinished = 0;

            for (int i = 0; i < numThreads; i++)
            {
                int start = i;  // note this is required, using i in the lambda expresssion fails (?!?)
                ThreadPool.QueueUserWorkItem(state => count(data, targets, result, start, numThreads));
            }

            // We rely on the workers incrementing an instance variable to signal they are done.
            while (countFinished < numThreads)
                Thread.Sleep(1);
        }

        // Launch a bunch of threads to do the comparison in parallel.
        // This version uses Tasks, a feature added to the thread pool in .NET 4.0.
        // Allows much more elegant way to wait for the thread pooled jobs to finish.
        public static void countTasks(int[] data, int[] targets, int[] result, int numThreads)
        {
            List<Task> tasks = new List<Task>();
            for (int i = 0; i < numThreads; i++)
            {
                int start = i;  // note this is required, using i in the lambda expresssion fails (?!?)
                Task t = Task.Run(() => count(data, targets, result, start, numThreads));
                tasks.Add(t);
            }

            Task.WaitAll(tasks.ToArray());
        }

        // Launch a bunch of threads to do the comparison in parallel.
        // This version makes use of the Parallel.For feature.
        // Parallel for was added to .NET v4.0.
        public static void countParallelFor(int[] data, int[] targets, int[] result, int numThreads)
        {
            Parallel.For(0, numThreads, i =>
            {
                count(data, targets, result, i, numThreads);
            });
        }

        // Compute some stats about a resulting array given a set of target values.
        // Used to verify the parallel versions all compute the same thing.
        static void stats(int[] result, int[] targets)
        {
            int maxValue = int.MinValue;
            int maxIndex = 0;
            int minValue = int.MaxValue;
            int minIndex = 0;
            double sum = 0.0;

            for (int i = 0; i < result.Length; i++)
            {
                if (result[i] > maxValue)
                {
                    maxValue = result[i];
                    maxIndex = i;
                }
                if (result[i] < minValue)
                {
                    minValue = result[i];
                    minIndex = i;
                }
                sum += result[i];
            }
            System.Console.WriteLine("Min occurences = {0}, value = {1}", minValue, targets[minIndex]);
            System.Console.WriteLine("Max occurences = {0}, value = {1}", maxValue, targets[maxIndex]);
            System.Console.WriteLine("Avg occurences = {0}", sum / result.Length);
        }

        // Allows the main program to define a list of methods to test
        public delegate void CountParallel(int[] data, int[] targets, int[] result, int numThreads);

        static void Main(string[] args)
        {
            // Create a random array with values in [0, 1000)
            int size = int.Parse(args[0]);
            int[] data = new int[size];
            Random rand = new Random();
            for (int i = 0; i < size; i++)
                data[i] = rand.Next(1000);

            // Create a set of target values we want to count up
            int numTargets = int.Parse(args[1]);
            int[] targets = new int[numTargets];

            for (int i = 0; i < numTargets; i++)
                targets[i] = rand.Next(1000);
            Stopwatch timer = new Stopwatch();

            const int MAX_THREADS = 8;

            // Serial version
            int[] result = new int[numTargets];           
            timer.Start();
            count(data, targets, result, 0, 1);
            timer.Stop();
            System.Console.WriteLine("Serial, ms = {0}", timer.ElapsedMilliseconds);
            //stats(result, targets);
            System.Console.WriteLine();

            List<CountParallel> methods = new List<CountParallel>();
            List<string>        names   = new List<string>();

            methods.Add(countThreaded);
            names.Add("Classic threaded");

            methods.Add(countPool);
            names.Add("Thread pool");

            methods.Add(countPool);
            names.Add("Thread tasks");

            methods.Add(countPool);
            names.Add("Parallel for");

            // Test out all the different parallel schemes
            for (int i = 0; i < methods.Count; i++)
            {
                // Test out a range of number of threads
                for (int numThreads = 2; numThreads <= MAX_THREADS; numThreads++)
                {
                    int[] result2 = new int[numTargets];
                    timer.Reset();
                    timer.Start();
                    methods[i](data, targets, result2, numThreads);
                    timer.Stop();
                    System.Console.WriteLine("{0}, {1} threads, ms = {2}", names[i], numThreads, timer.ElapsedMilliseconds);
                    //stats(result2, targets);                    
                }
                System.Console.WriteLine();
            }
        }
    }
}
